@@ -34,7 +34,6 @@ import time
 
 from perlcompat import warn, die
 import numpy
-import pytess
 import tbdump
 
 VERSION = 1.8
@@ -82,6 +81,8 @@ EXPORT_FORMATS = sorted(EXPORT_SUBP.keys())
 MAX_RETRIES = 100
 
 class Graph:
+    # FIXME: Support non-multiedged graph; multiedged option is always
+    # regarded as True.
     def __init__(self, directed=True, multiedged=True):
         self.G = {}  # Graph attributes.
         self.V = {}  # Vertices.
@@ -221,14 +222,18 @@ class Graph:
             return []
         return self.EO[u].keys()
 
-    def neighbors(self, v):
-        """Return all neighbor nodes of vetex V."""
-        found = set()
-        for u in self.predecessors(v, ignore=True):
-            found.add(u)
-        for u in self.successors(v, ignore=True):
-            found.add(u)
-        return found
+    def neighbors(self, v, k=1):
+        """Return all K-hop neighbor nodes of vetex V."""
+        if k <= 1:
+            in_vertices = set(self.predecessors(v, ignore=True))
+            out_vertices = set(self.successors(v, ignore=True))
+            return in_vertices.union(out_vertices)
+        else:
+            vertices = set(self.neighbors(v))
+            for u in self.neighbors(v):
+                found = self.neighbors(u, k - 1)
+                vertices.union(found)
+            return vertices
 
     def set_vertex_attribute(self, v, attr, val):
         """Define the vertex attribute of vetex V named ATTR as value VAL."""
@@ -288,6 +293,9 @@ class Graph:
         for v in alist:
             self.delete_vertex(v)
 
+    def delete_all_vertices(self):
+        self.delete_vertices(self.vertices())
+
     def random_vertex(self):
         """Randomly choose a vertex from all vertices."""
         return random.choice(list(self.vertices()))
@@ -343,7 +351,7 @@ class Graph:
         else:
             return 0
 
-    def add_edge(self, u, v):
+    def add_edge(self, u, v, weight=None):
         """Add an edge from vertex U to vertex V."""
         if self.undirected() and u > v:
             u, v = v, u
@@ -359,6 +367,8 @@ class Graph:
         if u not in self.EI[v]:
             self.EI[v][u] = {}
         self.EI[v][u][count] = {}  # default edge attributes
+        if weight:
+            self.set_edge_weight_by_id(u, v, count, weight)
         return count
 
     def delete_edge(self, u, v):
@@ -764,10 +774,13 @@ class Graph:
         object."""
         N = self.nvertices()
         m = numpy.zeros((N, N), int)
-        for ui, vi in self.edge_indices():
-            m[ui, vi] += 1
+        for u, v in self.edges():
+            w = self.get_edge_weight(u, v) or 1
+            ui = self.vertex_index(u)
+            vi = self.vertex_index(v)
+            m[ui, vi] += w
             if self.undirected():
-                m[vi, ui] += 1
+                m[vi, ui] += w
         return m
 
     def diagonal_matrix(self):
@@ -784,6 +797,17 @@ class Graph:
         object."""
         return self.diagonal_matrix() - self.adjacency_matrix()
 
+    def weight_matrix(self):
+        """Return the weight matrix of the graph."""
+        nvertices = self.nvertices()
+        m = numpy.zeros((nvertices, nvertices), int)
+        for u, v in self.edges():
+            w = self.get_edge_weight(u, v) or 1
+            ui = self.vertex_index(u)
+            vi = self.vertex_index(v)
+            m[ui][vi] = m[vi][ui] = w
+        return m
+
     def eigenvals(self, m):
         """Return the eigenvalues of matrix M.  Eigenvalues are sorted in the
         ascending order."""
@@ -860,19 +884,25 @@ class Graph:
     def merge_vertices(self, u, v):
         """Delete vertex U after connecting all neighbors (except vertex U) of
         vertex V to vertex U."""
+        if u == v:
+            return
         for t in self.neighbors(v):
-            if u == t:
+            if t == u:
                 continue
-            if not self.has_edge(u, t):
+            # Note: w might be None (i.e., no weight specified)
+            w = self.get_edge_weight(v, t)
+            if self.has_edge(u, t):
+                new_w = self.get_edge_weight(u, t) or 1
+                new_w += w or 1
+                self.set_edge_weight(u, t, new_w)
+            else:
                 self.add_edge(u, t)
-                w = self.get_edge_weight(v, t)
-                if w:
-                    self.set_edge_weight(u, t, w)
+                self.set_edge_weight(u, t, w)
             self.delete_edge(v, t)
         self.delete_edge(u, v)
         self.delete_vertex(v)
 
-    def random_matching_coarsening(self, alpha=.5):
+    def RM_coarsening(self, alpha=.5):
         # FIXME: Support directed graphs.
         self.expect_undirected()
         n = self.nvertices()
@@ -880,7 +910,7 @@ class Graph:
             u, v = self.random_edge()
             self.merge_vertices(u, v)
 
-    def coarsenet_coarsening(self, alpha=.5):
+    def COARSENET_coarsening(self, alpha=.5):
         def d_lambda(lmbda, u, v, a, b):
             # Weight of edges (a, b) and (b, a), respectively.
             # FIXME: This code assumes vertex IDs are 1...N.
@@ -895,7 +925,7 @@ class Graph:
             term4 = beta1 * u[b - 1] * v[a - 1]
             denom = numpy.dot(v.transpose(), u)
             delta = (-lmbda * term1 + term2 + term3 + term4) / (denom - term1)
-            return delta
+            return float(delta)
 
         # FIXME: Support directed graphs.
         self.expect_undirected()
@@ -912,10 +942,79 @@ class Graph:
         v = self.negate_if_negative(v)
         scores = [(d_lambda(lmbda, u, v, a, b), a, b) for a, b in self.edges()]
         scores = sorted(scores, key=lambda x: abs(x[0]))
+        for s, u, v in scores:
+            warn(f'coarsenet: score={s:6.3f} edge=({u}, {v})')
+
         while self.nvertices() > n * alpha:
             _, u, v = scores.pop(0)
             self.merge_vertices(u, v)
 
+    def MGC_coarsening(self, alpha=.5):
+        # FIXME: Seems to fail appropriate coarsening.
+        def find_vertices_with_minimum_distance():
+            adj = self.adjacency_matrix()
+            dmin = None
+            for v in self.vertices():
+                # Collect all 2-hop neighbors from vertex V.
+                neighbors = self.neighbors(v, 2)
+                dv = self.degree(v)
+                iv = self.vertex_index(v)
+                lv = adj[iv]
+                for u in neighbors:
+                    du = self.degree(u)
+                    iu = self.vertex_index(u)
+                    lu = adj[iu]
+                    d = numpy.linalg.norm(lv / dv - lu / du, ord=1)
+                    if dmin is None or d < dmin:
+                        dmin = d
+                        umin, vmin = u, v
+            return dmin, umin, vmin
+
+        # FIXME: Support directed graphs.
+        self.expect_undirected()
+        nvertices = self.nvertices()
+        while self.nvertices() > nvertices * alpha:
+            d, u, v = find_vertices_with_minimum_distance()
+            warn(f'MGC: d={d} edge=({u}, {v})')
+            self.merge_vertices(u, v)
+
+    def add_edges_from_sparse_graph(self, sg):
+        edges = sg.get_edge_list()
+        for ui, vi, w in zip(*edges):
+            u, v = ui + 1, vi + 1
+            self.add_edge(u, v)
+            self.set_edge_weight(u, v, w)
+
+    def local_variation_coarsening(self, alpha=.5, method='variation_neighborhoods'):
+        import pygsp
+        import graph_coarsening
+        # Create a sparse graph from the weight matrix.
+        g = pygsp.graphs.Graph(self.weight_matrix())
+        r = 1 - alpha  # r=0 means no reduction.
+        # C: coarsening matrix
+        # Gc: coarsed graph
+        # Call: coarsening matrix at all levels
+        # Gall: coarsed graph at all levels
+        C, Gc, Call, Gall = graph_coarsening.coarsen(g, r=r, method=method)
+        # Rewrite all edges according to Gc.
+        self.delete_vertices(list(self.vertices()))
+        self.add_edges_from_sparse_graph(Gc)
+
+    def LVN_coarsening(self, alpha=.5):
+        self.local_variation_coarsening(alpha=alpha, method='variation_neighborhood')
+
+    def LVE_coarsening(self, alpha=.5):
+        self.local_variation_coarsening(alpha=alpha, method='variation_edges')
+
+    def kron_coarsening(self, alpha=.5):
+        self.local_variation_coarsening(alpha=alpha, method='kron')
+
+    def HEM_coarsening(self, alpha=.5):
+        self.local_variation_coarsening(alpha=alpha, method='heavy_edge')
+
+    def affinity_coarsening(self, alpha=.5):
+        self.local_variation_coarsening(alpha=alpha, method='affinity_JC')
+
     # util ----------------------------------------------------------------
     def header_string(self, comment='# '):
         date = time.strftime('%Y/%M/%D %H:%M:%S', time.localtime())
@@ -1261,6 +1360,7 @@ class Graph:
         return self
 
     def create_voronoi_graph(self, npoints=10, width=1, height=1):
+        import pytess
         points = [(random.uniform(0, width), random.uniform(0, height))
                   for n in range(npoints)]
         polys = pytess.voronoi(points)
@@ -1622,11 +1722,17 @@ class Graph:
             l = '->' if self.is_directed() else '--'
             for i in self.get_multiedge_ids(u, v):
                 astr += f'  "{u}" {l} "{v}"'
+                # Dump edge attributes.
                 attrs = self.get_edge_attributes_by_id(u, v, i)
                 if attrs:
                     alist = []
                     for key, val in attrs.items():
-                        alist.append(f'{key}="{val}"')
+                        if val is None:
+                            continue
+                        if type(val) == str:
+                            alist.append(f'{key}="{val}"')
+                        else:
+                            alist.append(f'{key}={val}')
                     astr += ' [' + (', '.join(alist)) + ']'
                 astr += ';\n'
         astr += '}\n'
@@ -1657,7 +1763,8 @@ palette ecolor 0 .6 1 .5
             out += f'attach t{v} v{v}\n'
         n = 1
         for u, v in sorted(self.edges()):
-            out += f'define e{n} link v{u} v{v} 1 ecolor\n'
+            w = (self.get_edge_weight(u, v) or 1) * 2
+            out += f'define e{n} link v{u} v{v} {w} ecolor\n'
             n += 1
         out += 'spring /^v/\n'
         out += 'display\n'
